// #####################################################################

// Datalogger subsystem

#include "datalogger.h"
#include "sd.h"
#include "led.h"
#include "ntp.h"
#include "pmic.h"
#include "sensor.h"
#include "wifi.h"
#include "menu.h"
#include "ds18b20.h"
#include "main.h"

bool b_datalogger_enabled = false;
int datalogger_interval = 30;

unsigned long datalogger_log_millis;

const PROGMEM struct MENU_INTEGER_EXTRA extra_interval =
{
  .on_change = NULL,
  { .units = "s" },
  .minimum = 1,
  .maximum = 30 * 60,
  .increment = 1,
  .increment_fast = 5
};

const PROGMEM struct MENU_ENTRY menu_datalogger[] =
{
  { { .text = "Log Enabled" }, .type = MENU_BOOL, .data = (void *) &b_datalogger_enabled, .user_data = 0, .extra = NULL },
  { { .text = "Interval" }, .type = MENU_INTEGER, .data = (void *) &datalogger_interval, .user_data = 0, .extra = (void *) &extra_interval },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
unsigned long datalogger_routine_millis = 5000;
bool b_datalogger_initialized = false;

// #####################################################################

void datalogger_routine( )
{
  // check timer
  if( current_millis < datalogger_routine_millis )
    return;
  datalogger_routine_millis = current_millis + 1000;

  if( b_datalogger_enabled == false )
        return;

  if( b_datalogger_initialized == false && b_sd_initialized == true )
  {
    Serial.println(F("[Datalogger] Datalogger subsystem initialized."));
    b_datalogger_initialized = true;
  }
  else
  {
    if( current_millis > datalogger_log_millis )
    {
      datalogger_log_millis = current_millis + datalogger_interval * 1000;

      char file_path[64];

      // LED_IO_PIN indicates that currently input/output operation is going on SD card
      led_set_state( LED_IO_PIN, LED_ON );

      // Get time
      time_t epoch = rtc.getEpoch( );
      struct tm* tmp = gmtime( &epoch ); // struct tm is a structure that contains a calendar time broken down into its components

      // Check that all directories in file path exists
      strftime( file_path, sizeof(file_path), "%Y/%m", tmp );
      if( SD.exists( file_path ) == 0 )
      {
        if( SD.mkdir( file_path ) == 0 )
        {
          // Error: couldn't create directories
          led_set_state( LED_IO_PIN, LED_BLINK );
          return;
        }
      }

      // Make path to datalogging file
      strftime( file_path, sizeof(file_path), "%Y/%m/%Y%m%d.csv", tmp );

      // Open file for writting
      // By default position is at the end of file
      File datafile = SD.open( file_path, FILE_WRITE );
      if( datafile )
      {
        // Place header, if the file is new
        if( datafile.position( ) == 0 )
        {
          datafile.println(
            "\"Unix time\", "
            "\"Datetime\", "
            "\"USB voltage, V\", "
            "\"Battery voltage, V\", "
            "\"WiFi RSSI, dBm\", "
            "\"Internal temperature, °C\", "
            "\"Atmospheric pressure, kPa\", "
            "\"Humidity, % rH\", "
            "\"External temperature 1, °C\", "
            "\"External temperature 2, °C\", "
            "\"Digital output #1A\", "
            "\"Digital output #1B\", "
            "\"Digital output #2A\", "
            "\"Digital output #2B\", "
            "\"Digital output #3A\", "
            "\"Digital output #3B\""
          );
        }
        
        // Make Data/Time string
        char datetime_str[32];
        strftime( datetime_str, sizeof(datetime_str), "%Y-%m-%dT%H:%M:%S.000Z", tmp );

        // Make the logline
        char logline_str[256];
        String epoch_str = int64String( epoch );
        int logline_len = snprintf( logline_str, sizeof(logline_str), "\"%s\", \"%s\", \"%.3f\", \"%.3f\", \"%li\", \"%.3f\", \"%.3f\", \"%.3f\", \"%.3f\", \"%.3f\", \"%i\", \"%i\", \"%i\", \"%i\", \"%i\", \"%i\"\n",
          epoch_str.c_str( ), datetime_str,                     // timestamp
          usb_voltage, battery_voltage, (int) WiFi.RSSI( ),     // system values
          env_temperature, env_pressure, env_humidity,          // internal sensors
          ds18b20_sensor1, ds18b20_sensor2,                     // external sensors
          0, 0, 0, 0, 0, 0                                      // digital outputs
        );

        // Debugging RAM
        display_freeram( );
        Serial.println(logline_str);

        // Write the logline to file
        size_t written = datafile.write( (uint8_t *) &logline_str[0], logline_len );
        if( written != logline_len )
        {
          // Error: couldn't write the logline
          led_set_state( LED_IO_PIN, LED_BLINK );
          datafile.close( );
          return;
        }
        
        datafile.close( );
      }
      else
      {
        // Error: couldn't open datafile
        led_set_state( LED_IO_PIN, LED_BLINK );
        return;
      }

      led_set_state( LED_IO_PIN, LED_OFF );
    }
  }
}

// #####################################################################