// #####################################################################

// Keyboard variables

#include <Arduino.h>
#include <stdint.h>

#include "keyboard.h"
#include "menu.h"
#include "beep.h"
#include "lcd.h"
#include "gpio.h"

unsigned long key_esc_millis = 0;
unsigned long key_up_millis = 0;
unsigned long key_ok_millis = 0;
unsigned long key_left_millis = 0;
unsigned long key_down_millis = 0;
unsigned long key_right_millis = 0;

uint16_t buttons_prev = 0;
uint16_t buttons_current = 0;

int keyboard_special_command = 0;     // Counter for special command activation
unsigned long key_last_activity = 0;

extern unsigned long current_millis;
bool b_keyboard_initialized = false;
unsigned long keyboard_routine_millis = 50;

// #####################################################################

void keyboard_push_stack( )
{
  // Add previous menu to the stack together with positions, so that ESC could work as "Back" button
  if( menu_stack_position < sizeof(menu_stack) / sizeof(struct MENU_STACK) )
  {
    menu_stack[menu_stack_position].entry = menu_current;
    menu_stack[menu_stack_position].cursor_position = menu_cursor_position;
    menu_stack[menu_stack_position].screen_position = menu_screen_position;
    menu_stack_position++;
  }
}

void keyboard_pop_stack( )
{
  if( menu_current != NULL )
  {
    if( menu_stack_position > 0 )
    {
      // Return to the previous menu
      menu_current = menu_stack[menu_stack_position - 1].entry;
      menu_current_size = 0;
      menu_cursor_position = menu_stack[menu_stack_position - 1].cursor_position;
      menu_screen_position = menu_stack[menu_stack_position - 1].screen_position;
      menu_stack_position--;
      menu_stack[menu_stack_position].entry = NULL;
      menu_stack[menu_stack_position].cursor_position = 0;
      menu_stack[menu_stack_position].screen_position = 0;
    }
    else
    {
      // Close the menu and show the main screen
      menu_current = NULL;
      menu_current_size = 0;
    }
  }
}

// #####################################################################

void keyboard_event( unsigned char key, unsigned char flag )
{
  // Remember when the last time key was pressed
  // This can be used to turn off display after a set amount of time
  key_last_activity = current_millis;
  
  if( flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT )
    beep( 10 );
  
  // Holding ESC button long enough will trigger special command
  if( key == KEY_ESC && flag & KEYFLAG_REPEAT )
    keyboard_special_command++;
  else
    keyboard_special_command = 0;
  if( keyboard_special_command > 10 )
  {
    lcd.clear( );
    lcd.noCursor( );
  }

  if( key == KEY_ESC && flag & KEYFLAG_PRESS && menu_mode == MENU_NORMAL )
  {
    menu_horizontal_scroll = 0;
    keyboard_pop_stack( );
  }
  else if( key == KEY_OK && flag & KEYFLAG_PRESS )
  {
    if( menu_current == NULL )
    {
      // Open the main menu
      menu_current = (struct MENU_ENTRY *) menu_main;
      menu_current_size = 0;
    }
    else
    {
      //
      // Process action in the menu
      //

      switch( menu_current[menu_cursor_position].type )
      {
        case MENU_SUBMENU:
        {
          keyboard_push_stack( );

          // Call on_open() event if available
          if( menu_current[menu_cursor_position].extra )
            if( ((struct MENU_SUBMENU_EXTRA *) menu_current[menu_cursor_position].extra)->on_open )
              ((struct MENU_SUBMENU_EXTRA *) menu_current[menu_cursor_position].extra)->on_open( &menu_current[menu_cursor_position] );

          // Change active menu
          menu_current = (struct MENU_ENTRY *) menu_current[menu_cursor_position].data;
          menu_current_size = 0;

          // Reset cursor positions
          menu_cursor_position = 0;
          menu_screen_position = 0;
          break;
        }
        
        case MENU_COMMAND:
        {
          // Some hard core C hacking here
          //
          // The code below takes address to a function and type casts it
          // from (void *) to a function which can be invoked
          
          void (*function_ptr)(struct MENU_ENTRY *) = (void (*)(struct MENU_ENTRY *)) menu_current[menu_cursor_position].data;
          (*function_ptr)( &menu_current[menu_cursor_position] );
          break;
        }
        
        case MENU_INTEGER:
        {
          // Open "menu_edit_integer"
          strncpy( &menu_edit_integer[0].text[0], &menu_current[menu_cursor_position].text[0], LCD_WIDTH - 3 );
          menu_edit_integer[2].data = &menu_current[menu_cursor_position];
          keyboard_push_stack( );
          menu_cursor_position = 2;
          menu_screen_position = 0;
          menu_current = menu_edit_integer;
          menu_current_size = 0;

          // Set cursor to the lowest integer
          menu_integer_position = 0;
          break;
        }
        case MENU_EDIT_INTEGER:
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
          if( menu_mode == MENU_NORMAL )
          {
            // Display the cursor
            // It will indicate which number is being incremented or decremented
            //lcd.cursor( );

            // Save original value in case user cancels editing
            menu_original_integer = *((int *) menu_entry[0].data);
            
            menu_mode = MENU_EDITING_INTEGER;
          }
          else
          {
            *((int *) menu_entry[0].data) = menu_original_integer;
            menu_mode = MENU_NORMAL;
          }
          break;
        }

        case MENU_BOOL:
        {
          // Open "menu_edit_bool"
          strncpy( &menu_edit_bool[0].text[0], &menu_current[menu_cursor_position].text[0], LCD_WIDTH - 3 );
          menu_edit_bool[2].data = &menu_current[menu_cursor_position];
          keyboard_push_stack( );
          menu_cursor_position = 2;
          menu_screen_position = 0;
          menu_current = menu_edit_bool;
          menu_current_size = 0;
          break;
        }
        case MENU_EDIT_BOOL:
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
          if( menu_mode == MENU_NORMAL )
          {
            // Save original value in case user cancels editing
            menu_original_bool = *((bool *) menu_entry[0].data);
            menu_mode = MENU_EDITING_BOOL;
          }
          else
          {
            *((bool *) menu_entry[0].data) = menu_original_bool;
            menu_mode = MENU_NORMAL;
          }
          break;
        }

        case MENU_FLOAT:
        {
          // Open "menu_edit_float"
          // Copy the name of the variable into float editing menu
          // and copy pointer to MENU_ENTRY of edited variable
          strncpy( &menu_edit_float[0].text[0], &menu_current[menu_cursor_position].text[0], LCD_WIDTH - 3 );
          menu_edit_float[2].data = &menu_current[menu_cursor_position];
          keyboard_push_stack( );
          menu_cursor_position = 2;
          menu_screen_position = 0;
          menu_current = menu_edit_float;
          menu_current_size = 0;

          // Set cursor to the lowest integer
          menu_float_position = 0;
          break;
        }
        case MENU_EDIT_FLOAT:
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
          if( menu_mode == MENU_NORMAL )
          {
            // Save original value in case user cancels editing
            menu_original_float = *((float *) menu_entry[0].data);
            menu_mode = MENU_EDITING_FLOAT;
          }
          else
          {
            *((float *) menu_entry[0].data) = menu_original_float;
            menu_mode = MENU_NORMAL;
          }
          break;
        }

        case MENU_TEXT:
        {
          // Open "menu_edit_text"
          // Copy the name of the variable into float editing menu
          // and copy pointer to MENU_ENTRY of edited variable
          strncpy( &menu_edit_text[0].text[0], &menu_current[menu_cursor_position].text[0], LCD_WIDTH - 3 );
          menu_edit_text[2].data = &menu_current[menu_cursor_position];
          keyboard_push_stack( );
          menu_cursor_position = 2;
          menu_screen_position = 0;
          menu_current = menu_edit_text;
          menu_current_size = 0;

          // Set cursor to the lowest integer
          menu_text_position = 0;
          menu_text_scroll = 0;
          break;
        }
        case MENU_EDIT_TEXT:
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
          if( menu_mode == MENU_NORMAL )
          {
            // Save original value in case user cancels editing
            strncpy( menu_original_text, (char *) menu_entry[0].data, sizeof(menu_original_text) );
            menu_mode = MENU_EDITING_TEXT;
          }
          else
          {
            strncpy( (char *) menu_entry[0].data, menu_original_text, ((struct MENU_TEXT_EXTRA *) menu_entry[0].extra)->text_size );
            menu_mode = MENU_NORMAL;
          }
          break;
        }
      }
    }
  }
  else if( key == KEY_ESC && flag & KEYFLAG_PRESS && menu_mode == MENU_EDITING_INTEGER )
  {
    // Hide the cursor
    //lcd.noCursor( );

    // Restore original integer value
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    // *((int *) menu_entry[0].data) = menu_original_integer;
    
    menu_mode = MENU_NORMAL;
  }
  else if( key == KEY_ESC && flag & KEYFLAG_PRESS && menu_mode == MENU_EDITING_BOOL )
  {
    // Restore original bool value
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    //*((bool *) menu_entry[0].data) = menu_original_bool;
    
    menu_mode = MENU_NORMAL;
  }
  else if( key == KEY_ESC && flag & KEYFLAG_PRESS && menu_mode == MENU_EDITING_FLOAT )
  {
    // Restore original float value
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    //*((float *) menu_entry[0].data) = menu_original_float;
    
    menu_mode = MENU_NORMAL;
  }
  else if( key == KEY_ESC && flag & KEYFLAG_PRESS && menu_mode == MENU_EDITING_TEXT )
  {
    // Restore original text
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    //strncpy( (char *) menu_entry[0].data, menu_original_text, menu_entry[0].text_size );
    
    menu_mode = MENU_NORMAL;
  }

  if( menu_current != NULL && menu_current_size != 0 && menu_mode == MENU_NORMAL )
  {
    if( key == KEY_UP && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_cursor_position > 0 )
      {
        menu_cursor_position -= 1;
        if( menu_screen_position > menu_cursor_position )
          menu_screen_position--;
      }
      else
      {
        // Jump to the end of the menu
        menu_cursor_position = menu_current_size - 1;
        if( menu_current_size >= LCD_HEIGHT )
          menu_screen_position = menu_current_size - LCD_HEIGHT;
      }

      menu_horizontal_scroll = 0;
    }
    else if( key == KEY_DOWN && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_cursor_position + 1 < menu_current_size )
      {
        menu_cursor_position += 1;
        if( menu_screen_position + LCD_HEIGHT < menu_cursor_position + 1 )
          menu_screen_position++;
      } 
      else
      {
        // Jump to the start of the menu
        menu_cursor_position = 0;
        menu_screen_position = 0;
      }

      menu_horizontal_scroll = 0;
    }
    else if( key == KEY_LEFT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_horizontal_scroll > 0 )
        menu_horizontal_scroll--;
    }
    else if( key == KEY_RIGHT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_horizontal_scroll < 256 )
        menu_horizontal_scroll++;
    }
  }

  if( menu_mode == MENU_EDITING_INTEGER )
  {
    // Calculate increment depending on menu_integer_position
    // basically: factor = 10^menu_integer_position
    int factor = 1;
    for( int i = 0; menu_integer_position > i; i++ )
      factor *= 10;
    
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    int *value = &menu_original_integer; //(int *) menu_entry[0].data;
    if( key == KEY_UP && flag & KEYFLAG_PRESS )
    {
      if( *value + ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment * factor > ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->maximum )
        *value = ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->maximum;
      else
        *value += ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment * factor;
    }
    else if( key == KEY_DOWN && flag & KEYFLAG_PRESS )
    {
      if( *value - ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment * factor < ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->minimum )
        *value = ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->minimum;
      else
        *value -= ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment * factor;
    }
    else if( key == KEY_UP && flag & KEYFLAG_REPEAT )
    {
      if( *value + ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment_fast * factor > ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->maximum )
        *value = ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->maximum;
      else
        *value += ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment_fast * factor;
    }
    else if( key == KEY_DOWN && flag & KEYFLAG_REPEAT )
    {
      if( *value - ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment_fast * factor < ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->minimum )
        *value = ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->minimum;
      else
        *value -= ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->increment_fast * factor;
    }
    else if( key == KEY_LEFT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_integer_position < 9 )
        menu_integer_position++;
    }
    else if( key == KEY_RIGHT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_integer_position > 0 )
        menu_integer_position--;
    }

    // Call on_change() event if available
    if( key == KEY_UP || key == KEY_DOWN )
    {
      if( menu_entry[0].extra )
        if( ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->on_change )
          ((struct MENU_INTEGER_EXTRA *) menu_entry[0].extra)->on_change( menu_entry );
    }
  }
  else if( menu_mode == MENU_EDITING_BOOL )
  {
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    //bool *value = (bool *) menu_entry[0].data;
    if( (key == KEY_UP || key == KEY_DOWN) && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
        //*value = !(*value);
        menu_original_bool = !menu_original_bool;
    }
  }
  else if( menu_mode == MENU_EDITING_FLOAT )
  {
    // Calculate increment depending on menu_integer_position
    // basically: factor = 10^menu_integer_position
    float factor = pow( 10.0, (float) menu_float_position );

    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    float *value = &menu_original_float; //(float *) menu_entry[0].data;
    if( key == KEY_UP && flag & KEYFLAG_PRESS )
    {
      if( *value + ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment * factor > ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_maximum )
        *value = ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_maximum;
      else
        *value += ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment * factor;
    }
    else if( key == KEY_DOWN && flag & KEYFLAG_PRESS )
    {
      if( *value - ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment * factor < ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_minimum )
        *value = ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_minimum;
      else
        *value -= ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment * factor;
    }
    else if( key == KEY_UP && flag & KEYFLAG_REPEAT )
    {
      if( *value + ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment_fast * factor > ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_maximum )
        *value = ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_maximum;
      else
        *value += ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment_fast * factor;
    }
    else if( key == KEY_DOWN && flag & KEYFLAG_REPEAT )
    {
      if( *value - ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment_fast * factor < ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_minimum )
        *value = ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_minimum;
      else
        *value -= ((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->f_increment_fast * factor;
    }
    else if( key == KEY_LEFT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_float_position < 9 )
        menu_float_position++;
    }
    else if( key == KEY_RIGHT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_float_position > -((struct MENU_FLOAT_EXTRA *) menu_entry[0].extra)->precision )
        menu_float_position--;
    }
  }
  else if( menu_mode == MENU_EDITING_TEXT )
  {
    struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[2].data;
    char *text_buf = menu_original_text; //(char *) menu_entry[0].data;
    size_t text_len = strlen( text_buf );
    if( key == KEY_UP && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( text_buf[menu_text_position] == 0 )
        text_buf[menu_text_position] = 0x020;
      else if( text_buf[menu_text_position] < 0x07F )
        text_buf[menu_text_position]++;
    }
    else if( key == KEY_DOWN && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( text_buf[menu_text_position] > 0x020 )
        text_buf[menu_text_position]--;
      else
        text_buf[menu_text_position] = 0;
    }
    else if( key == KEY_LEFT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_text_position > 0 )
        menu_text_position--;

      if( menu_text_scroll > menu_text_position )
          menu_text_scroll--;
    }
    else if( key == KEY_RIGHT && (flag & KEYFLAG_PRESS || flag & KEYFLAG_REPEAT) )
    {
      if( menu_text_position + 1 < ((struct MENU_TEXT_EXTRA *) menu_entry[0].extra)->text_size )
        menu_text_position++;

      if( menu_text_scroll + LCD_WIDTH - 3 < menu_text_position + 1 )
        menu_text_scroll++;
    }
  }
  
  //
  // Code to test button press
  //
  /*if( flag & KEYFLAG_PRESS )
  switch( key )
  {
    case KEY_ESC: Serial.println(F("[Keyboard] KEY_ESC pressed.")); break;
    case KEY_UP: Serial.println(F("[Keyboard] KEY_UP pressed.")); break;
    case KEY_OK: Serial.println(F("[Keyboard] KEY_OK pressed.")); break;
    case KEY_LEFT: Serial.println(F("[Keyboard] KEY_LEFT pressed.")); break;
    case KEY_DOWN: Serial.println(F("[Keyboard] KEY_DOWN pressed.")); break;
    case KEY_RIGHT: Serial.println(F("[Keyboard] KEY_RIGHT pressed.")); break;
  }

  if( flag & KEYFLAG_RELEASE )
  switch( key )
  {
    case KEY_ESC: Serial.println(F("[Keyboard] KEY_ESC released.")); break;
    case KEY_UP: Serial.println(F("[Keyboard] KEY_UP released.")); break;
    case KEY_OK: Serial.println(F("[Keyboard] KEY_OK released.")); break;
    case KEY_LEFT: Serial.println(F("[Keyboard] KEY_LEFT released.")); break;
    case KEY_DOWN: Serial.println(F("[Keyboard] KEY_DOWN released.")); break;
    case KEY_RIGHT: Serial.println(F("[Keyboard] KEY_RIGHT released.")); break;
  }*/
}

// #####################################################################

void keyboard_routine( )
{
  // check timer
  if( current_millis < keyboard_routine_millis )
    return;
  keyboard_routine_millis = current_millis + 20;

  if( b_keyboard_initialized == false )
  {
    Serial.println(F("[Keyboard] Keyboard initialized."));
    b_keyboard_initialized = true;
  }
  else
  {
    if( b_gpio_initialized == true )
    {
      uint16_t buttons_current = PCF.readButton16( );

      // Compare current and previous states to detect key press or release
      if( !(buttons_prev & 0b0000001) && (buttons_current & 0b0000001) )   { keyboard_event( KEY_ESC, KEYFLAG_PRESS );    key_esc_millis = current_millis + 250; }
      if( !(buttons_prev & 0b0000010) && (buttons_current & 0b0000010) )   { keyboard_event( KEY_UP, KEYFLAG_PRESS );     key_up_millis = current_millis + 250; }
      if( !(buttons_prev & 0b0000100) && (buttons_current & 0b0000100) )   { keyboard_event( KEY_OK, KEYFLAG_PRESS );     key_ok_millis = current_millis + 250; }
      if( !(buttons_prev & 0b0001000) && (buttons_current & 0b0001000) )   { keyboard_event( KEY_LEFT, KEYFLAG_PRESS );   key_left_millis = current_millis + 250; }
      if( !(buttons_prev & 0b0010000) && (buttons_current & 0b0010000) )   { keyboard_event( KEY_DOWN, KEYFLAG_PRESS );   key_down_millis = current_millis + 250; }
      if( !(buttons_prev & 0b0100000) && (buttons_current & 0b0100000) )   { keyboard_event( KEY_RIGHT, KEYFLAG_PRESS );  key_right_millis = current_millis + 250; }
      if( (buttons_prev & 0b0000001) && !(buttons_current & 0b0000001) )   keyboard_event( KEY_ESC, KEYFLAG_RELEASE );
      if( (buttons_prev & 0b0000010) && !(buttons_current & 0b0000010) )   keyboard_event( KEY_UP, KEYFLAG_RELEASE );
      if( (buttons_prev & 0b0000100) && !(buttons_current & 0b0000100) )   keyboard_event( KEY_OK, KEYFLAG_RELEASE );
      if( (buttons_prev & 0b0001000) && !(buttons_current & 0b0001000) )   keyboard_event( KEY_LEFT, KEYFLAG_RELEASE );
      if( (buttons_prev & 0b0010000) && !(buttons_current & 0b0010000) )   keyboard_event( KEY_DOWN, KEYFLAG_RELEASE );
      if( (buttons_prev & 0b0100000) && !(buttons_current & 0b0100000) )   keyboard_event( KEY_RIGHT, KEYFLAG_RELEASE );

      if( (buttons_current & 0b0000001) && key_esc_millis < current_millis )    { keyboard_event( KEY_ESC, KEYFLAG_REPEAT );    key_esc_millis = current_millis + 100; }
      if( (buttons_current & 0b0000010) && key_up_millis < current_millis )     { keyboard_event( KEY_UP, KEYFLAG_REPEAT );     key_up_millis = current_millis + 100; }
      if( (buttons_current & 0b0000100) && key_ok_millis < current_millis )     { keyboard_event( KEY_OK, KEYFLAG_REPEAT );     key_ok_millis = current_millis + 100; }
      if( (buttons_current & 0b0001000) && key_left_millis < current_millis )   { keyboard_event( KEY_LEFT, KEYFLAG_REPEAT );   key_left_millis = current_millis + 100; }
      if( (buttons_current & 0b0010000) && key_down_millis < current_millis )   { keyboard_event( KEY_DOWN, KEYFLAG_REPEAT );   key_down_millis = current_millis + 100; }
      if( (buttons_current & 0b0100000) && key_right_millis < current_millis )  { keyboard_event( KEY_RIGHT, KEYFLAG_REPEAT );  key_right_millis = current_millis + 100; }

      buttons_prev = buttons_current;
    }
  }
}

// #####################################################################
