// #####################################################################

// LED subsystem

#include "led.h"
#include "gpio.h"
#include "keyboard.h"
#include "menu.h"

enum LED_STATE led_power = LED_OFF;
enum LED_STATE led_timesync = LED_OFF;
enum LED_STATE led_wifi = LED_OFF;
enum LED_STATE led_io = LED_OFF;

uint16_t led_gpio_state = 0;

int led_auto_off = 300;
bool led_always_on = false;
bool led_off = false;

const PROGMEM struct MENU_INTEGER_EXTRA extra_timeout =
{
  .on_change = NULL,
  { .units = "s" },
  .minimum = 0,
  .maximum = 3600,
  .increment = 1,
  .increment_fast = 5
};

const PROGMEM struct MENU_ENTRY menu_led[] =
{
  { { .text = "Auto-off" }, .type = MENU_INTEGER, .data = (void *) &led_auto_off, .user_data = 0, .extra = (void *) &extra_timeout },
  { { .text = "Always-On" }, .type = MENU_BOOL, .data = &led_always_on, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
bool b_led_initialized = false;
unsigned long led_routine_millis = 100;

// #####################################################################

void led_set_state( unsigned char pin, enum LED_STATE led_state )
{
  // Use this function to update state of a LED
  switch( pin )
  {
    case LED_POWER_PIN:       led_power = led_state; break;
    case LED_TIMESYNC_PIN:    led_timesync = led_state; break;
    case LED_WIFI_PIN:        led_wifi = led_state; break;
    case LED_IO_PIN:          led_io = led_state; break;
  }

  // Force output update
  led_routine_millis = 0;
  led_routine( );
}

// #####################################################################

void led_routine( )
{
  // check timer
  if( current_millis < led_routine_millis )
    return;
  led_routine_millis = current_millis + 100;

  if( b_led_initialized == false )
  {
    if( b_gpio_initialized == true )
    {
      // When Arduino is reprogrammed, the GPIO still remembers state,
      // so we reset it
      PCF.write( LED_POWER_PIN, HIGH );
      PCF.write( LED_TIMESYNC_PIN, HIGH );
      PCF.write( LED_WIFI_PIN, HIGH );
      PCF.write( LED_IO_PIN, HIGH );
      
      Serial.println(F("[LED] LED subsystem initialized."));
      b_led_initialized = true;
    }
  }
  else
  {
    if( b_gpio_initialized == true )
    {
      // Automatically turn off all leds to preserve power
      if( key_last_activity + (1000 * led_auto_off) < current_millis && led_always_on == false )
      {
        if( led_off == false )
        {
          PCF.write( LED_POWER_PIN, HIGH );
          PCF.write( LED_TIMESYNC_PIN, HIGH );
          PCF.write( LED_WIFI_PIN, HIGH );
          PCF.write( LED_IO_PIN, HIGH );
          led_gpio_state =
            (1 << LED_POWER_PIN) |
            (1 << LED_TIMESYNC_PIN) |
            (1 << LED_WIFI_PIN) |
            (1 << LED_IO_PIN);
          led_off = true;
        }
        return;
      }
      led_off = false;
      
      // Here I try to write code that only send state change to GPIO
      // instead of spamming with messages every time

      // Output is inversed. LOW GPIO output means LED is ON.
      if( (led_gpio_state & (1 << LED_POWER_PIN)) && (led_power == LED_ON) )           PCF.write( LED_POWER_PIN, LOW );
      if( (led_gpio_state & (1 << LED_TIMESYNC_PIN)) && (led_timesync == LED_ON) )     PCF.write( LED_TIMESYNC_PIN, LOW );
      if( (led_gpio_state & (1 << LED_WIFI_PIN)) && (led_wifi == LED_ON) )             PCF.write( LED_WIFI_PIN, LOW );
      if( (led_gpio_state & (1 << LED_IO_PIN)) && (led_io == LED_ON) )                 PCF.write( LED_IO_PIN, LOW );
      if( !(led_gpio_state & (1 << LED_POWER_PIN)) && (led_power == LED_OFF) )         PCF.write( LED_POWER_PIN, HIGH );
      if( !(led_gpio_state & (1 << LED_TIMESYNC_PIN)) && (led_timesync == LED_OFF) )   PCF.write( LED_TIMESYNC_PIN, HIGH );
      if( !(led_gpio_state & (1 << LED_WIFI_PIN)) && (led_wifi == LED_OFF) )           PCF.write( LED_WIFI_PIN, HIGH );
      if( !(led_gpio_state & (1 << LED_IO_PIN)) && (led_io == LED_OFF) )               PCF.write( LED_IO_PIN, HIGH );

      // Blinking implementation
      if( led_power == LED_BLINK )       PCF.write( LED_POWER_PIN, ( (led_routine_millis % 1000) > 500 ? LOW : HIGH ) );
      if( led_timesync == LED_BLINK )    PCF.write( LED_TIMESYNC_PIN, ( (led_routine_millis % 1000) > 500 ? LOW : HIGH ) );
      if( led_wifi == LED_BLINK )        PCF.write( LED_WIFI_PIN, ( (led_routine_millis % 1000) > 500 ? LOW : HIGH ) );
      if( led_io == LED_BLINK )          PCF.write( LED_IO_PIN, ( (led_routine_millis % 1000) > 500 ? LOW : HIGH ) );

      led_gpio_state = PCF.read16( );
    }
  }
}

// #####################################################################
