// #####################################################################

/*

  PIK-PIK RTU2

  created 22.04.2022

  by Valentin But <valentin.but@eesti.ee>

*/

// #####################################################################

// Power consumption:
//
// LCD    Backlight   LEDs  Wifi  Current
// 1      1           0     0     178 mA
// 1      0           0     0     55 mA
// 1      1           1     1     185 ... 210 mA
// 0      0           0     1     53 ... 82 mA

// #####################################################################

// Timer set up

// Generally, you should use "unsigned long" for variables that hold time
// The value will quickly become too large for an int to store

unsigned long current_millis = 0;

// #####################################################################

#include "arduino_secrets.h"

#include "serial.h"
#include "datalogger.h"
#include "ntp.h"
#include "watchdog.h"
#include "wifi.h"
#include "webserver.h"
#include "tunnel.h"
#include "sensor.h"
#include "ds18b20.h"
#include "gpio.h"
#include "lcd.h"
#include "keyboard.h"
#include "led.h"
#include "sd.h"
#include "pmic.h"
#include "beep.h"
#include "menu.h"

// #####################################################################

void factory_default( struct MENU_ENTRY *menu_entry )
{
  lcd_color[0] = 255;
  lcd_color[1] = 255;
  lcd_color[2] = 255;
  lcd_auto_powersave = 60;
  lcd_auto_off = 300;
  lcd_always_on = false;
  lcd_animation = false;

  pmic_input_current_limit = 2.0;
  pmic_input_voltageLimit = 3.88;
  pmic_minimum_system_voltage = 3.5;
  pmic_charge_voltage = 4.2;
  pmic_charge_current = 0.375;

  b_wifi_autoconnect = false;
  memset( wifi_networks, 0, sizeof(wifi_networks) );

  watchdog_enabled = true;

  sensor_max_attempts = 1;

  led_auto_off = 5 * 60;

  beep_enabled = true;

  b_rtc_auto_sync = true;
  rtc_timezone = 3;

  if( menu_entry != NULL )    // Display message only when called from menu
  {
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "Settings resseted", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "to factory default", LCD_WIDTH + 1 );
    menu_display_messagebox( );
  }
}

// #####################################################################

extern "C" char* sbrk( int incr );
int freeRam2( )
{
  char top;
  return &top - reinterpret_cast<char*>( sbrk( 0 ) );
}

void display_freeram( )
{
  Serial.print( F("- SRAM left: ") );
  Serial.println( freeRam2( ) );
}

// #####################################################################

void restart( struct MENU_ENTRY *menu_entry )
{
  NVIC_SystemReset( );
}

// #####################################################################

void setup( )
{
  factory_default( NULL );

  // Temporary quick connect
  // First attempt always fails
  wifi_status = WiFi.begin( SECRET_SSID, SECRET_PASS );
  delay( 1000 );
  wifi_status = WiFi.begin( SECRET_SSID, SECRET_PASS );
}

// #####################################################################

void update_millis( )
{
  unsigned long prev_millis = current_millis;
  current_millis = millis( );
  if( current_millis < prev_millis )
  {
    // millis overflow detected, we have to reset all the timers
    datalogger_routine_millis = 1000;
    watchdog_routine_millis = 1000;
    sd_routine_millis = 1000;
    beep_routine_millis = 50;
    menu_routine_millis = 250;
    led_routine_millis = 100;
    keyboard_routine_millis = 50;
    serial_routine_millis = 20;
    lcd_routine_millis = 500;
    wifi_routine_millis = 1000;
    wifi_connect_millis = 2000;
    ntp_routine_millis = 5000;
    webserver_routine_millis = 1000;
    tunnel_routine_millis = 1500;
    sensor_routine_millis = 1000;
    gpio_routine_millis = 5000;
  }
}

// #####################################################################

void loop( ) {
  update_millis( );

  ds18b20_routine( );
  datalogger_routine( );
  watchdog_routine( );
  serial_routine( );
  pmic_routine( );
  sd_routine( );
  beep_routine( );
  menu_routine( );
  led_routine( );
  sensor_routine( );
  keyboard_routine( );
  wifi_routine( );
  ntp_routine( );
  webserver_routine( );
  tunnel_routine( );
  gpio_routine( );
  lcd_routine( );
}

// #####################################################################

// This is a dirty hack to keep arduino alive between read and write calls in a loop
void loop_io( )
{
  update_millis( );

  //ds18b20_routine( );
  watchdog_routine( );
  //serial_routine( );
  //pmic_routine( );
  //beep_routine( );
  //menu_routine( );
  //led_routine( );
  //sensor_routine( );
  //keyboard_routine( );
  //gpio_routine( );
  //lcd_routine( );
}

// #####################################################################

void test( struct MENU_ENTRY *menu_entry )
{
      // Get time
      time_t epoch = rtc.getEpoch( );
      String epoch_str = int64String(epoch);
      
      // Make the logline
      char logline_str[256];
      int logline_len = snprintf( logline_str, sizeof(logline_str), "\"%s\"", epoch_str.c_str() );

      Serial.print( logline_str );
}

// #####################################################################
