// #####################################################################

// SD card subsystem

#include "sd.h"
#include "menu.h"
#include "wifi.h"
#include "datalogger.h"
#include "pmic.h"
#include "watchdog.h"
#include "sensor.h"
#include "led.h"
#include "beep.h"
#include "ntp.h"

// set up variables using the SD utility library functions:
Sd2Card card;
SdVolume volume;
SdFile root;

// change this to match your SD shield or module;
// Arduino Ethernet shield: pin 4
// Adafruit SD shields and modules: pin 10
// Sparkfun SD shield: pin 8
// MKRZero SD: SDCARD_SS_PIN
const int chipSelect = 4;

enum SD_STATUS sd_status = SD_STATUS_NOT_INITIALIZED;

uint32_t volumesize;
char volumesize_txt[16];

struct MENU_ENTRY menu_sd[] =
{
  { { .text = "SD status: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Card type: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Filesystem type: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Size: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
unsigned long sd_routine_millis = 1000;
bool b_sd_initialized = false;

// #####################################################################

struct SD_BINARY_CONFIGURATION
{
  uint8_t version;

  // LCD
  int lcd_color[3];
  int lcd_auto_powersave;
  int lcd_auto_off;
  bool lcd_always_on;
  bool reserv1;

  // Battery
  float pmic_input_current_limit;
  float pmic_input_voltageLimit;
  float pmic_minimum_system_voltage;
  float pmic_charge_voltage;
  float pmic_charge_current;

  // WiFi
  bool b_wifi_autoconnect;
  struct WIFI_NETWORK wifi_networks[10];

  // Watchdog
  bool watchdog_enabled;

  // ENV Shield
  int sensor_max_attempts;

  // LED
  int led_auto_off;

  // Beep
  bool beep_enabled;

  // RTC
  int rtc_timezone;
  bool b_rtc_auto_sync;

  // Datalogger
  bool b_datalogger_enabled;
  int datalogger_interval;

  // LCD
  int lcd_animation;

  // LED
  bool led_always_on;

  // GPIO
  struct GPIO_DIGITAL_OUTPUT gpio_digital_outputs[GPIO_DIGITAL_OUTPUTS];
}
sd_configuration;

// #####################################################################

void sd_store_configuration_command( struct MENU_ENTRY *menu_entry )
{
  File myFile;
  
  if( b_sd_initialized != true )
    goto fail;
  
  // Transfer settings to the structure
  // TODO: Maybe use this structure as settings container instead of transferring each time?
  sd_configuration.version = 1;
  
  sd_configuration.lcd_color[0] = lcd_color[0];
  sd_configuration.lcd_color[1] = lcd_color[1];
  sd_configuration.lcd_color[2] = lcd_color[2];
  sd_configuration.lcd_auto_powersave = lcd_auto_powersave;
  sd_configuration.lcd_auto_off = lcd_auto_off;
  sd_configuration.lcd_always_on = lcd_always_on;
  sd_configuration.lcd_animation = lcd_animation;

  sd_configuration.pmic_input_current_limit = pmic_input_current_limit;
  sd_configuration.pmic_input_voltageLimit = pmic_input_voltageLimit;
  sd_configuration.pmic_minimum_system_voltage = pmic_minimum_system_voltage;
  sd_configuration.pmic_charge_voltage = pmic_charge_voltage;
  sd_configuration.pmic_charge_current = pmic_charge_current;

  sd_configuration.b_wifi_autoconnect = b_wifi_autoconnect;
  memcpy( &sd_configuration.wifi_networks, wifi_networks, sizeof(wifi_networks) );

  sd_configuration.watchdog_enabled = watchdog_enabled;

  sd_configuration.sensor_max_attempts = sensor_max_attempts;

  sd_configuration.led_auto_off = led_auto_off;
  sd_configuration.led_always_on = led_always_on;

  sd_configuration.beep_enabled = beep_enabled;
  
  sd_configuration.rtc_timezone = rtc_timezone;
  sd_configuration.b_rtc_auto_sync = b_rtc_auto_sync;

  sd_configuration.b_datalogger_enabled = b_datalogger_enabled;
  sd_configuration.datalogger_interval = datalogger_interval;

  memcpy( &sd_configuration.gpio_digital_outputs, gpio_digital_outputs, sizeof(gpio_digital_outputs) );

  myFile = SD.open( "conf.bin", FILE_WRITE | O_TRUNC );
  if( myFile )
  {
    //myFile.seek( 0 );
    size_t written = myFile.write( (uint8_t *) &sd_configuration, sizeof(struct SD_BINARY_CONFIGURATION) );
    myFile.close( );

    if( written != sizeof(struct SD_BINARY_CONFIGURATION) )
      goto fail;
    
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "Configuration st-", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "ored successfully", LCD_WIDTH + 1 );
  }
  else
  {
 fail:
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Error -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "Configuration st-", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "ore failed!", LCD_WIDTH + 1 );
  }

  menu_display_messagebox( );
}

// #####################################################################

void sd_read_configuration_command( struct MENU_ENTRY *menu_entry )
{
  File myFile;
  
  if( b_sd_initialized != true )
    goto fail;
  
  myFile = SD.open( "conf.bin", FILE_READ );
  if( myFile )
  {
    // Fill configuration with zeroes, so after read it is possible to check if read was successfull
    memset( &sd_configuration, 0, sizeof(struct SD_BINARY_CONFIGURATION) );
    
    myFile.read( (uint8_t *) &sd_configuration, sizeof(struct SD_BINARY_CONFIGURATION) );
    myFile.close( );

    // Check version
    if( sd_configuration.version != 1 )
      goto fail;

    lcd_color[0] = sd_configuration.lcd_color[0];
    lcd_color[1] = sd_configuration.lcd_color[1];
    lcd_color[2] = sd_configuration.lcd_color[2];
    lcd_auto_powersave = sd_configuration.lcd_auto_powersave;
    lcd_auto_off = sd_configuration.lcd_auto_off;
    lcd_always_on = sd_configuration.lcd_always_on;
    lcd_animation = sd_configuration.lcd_animation;
    
    pmic_input_current_limit = sd_configuration.pmic_input_current_limit;
    pmic_input_voltageLimit = sd_configuration.pmic_input_voltageLimit;
    pmic_minimum_system_voltage = sd_configuration.pmic_minimum_system_voltage;
    pmic_charge_voltage = sd_configuration.pmic_charge_voltage;
    pmic_charge_current = sd_configuration.pmic_charge_current; 

    b_wifi_autoconnect = sd_configuration.b_wifi_autoconnect;
    memcpy( wifi_networks, &sd_configuration.wifi_networks, sizeof(wifi_networks) );

    watchdog_enabled = sd_configuration.watchdog_enabled;

    sensor_max_attempts = sd_configuration.sensor_max_attempts;

    led_auto_off = sd_configuration.led_auto_off;
    led_always_on = sd_configuration.led_always_on;

    beep_enabled = sd_configuration.beep_enabled;

    rtc_timezone = sd_configuration.rtc_timezone;
    b_rtc_auto_sync = sd_configuration.b_rtc_auto_sync;
    
    b_datalogger_enabled = sd_configuration.b_datalogger_enabled;
    datalogger_interval = sd_configuration.datalogger_interval;

    memcpy( gpio_digital_outputs, &sd_configuration.gpio_digital_outputs, sizeof(gpio_digital_outputs) );

    // Do not show messagebox when invoked outside menu
    // This is for situation when the device is powered on
    if( menu_entry != NULL )
    {
      //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
      strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
      strncpy( menu_messagebox[1].text, "Configuration re-", LCD_WIDTH + 1 );
      strncpy( menu_messagebox[2].text, "ad successfull", LCD_WIDTH + 1 );
      menu_display_messagebox( );
    }
  }
  else
  {
 fail:
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Error -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "Configuration re-", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "ad failed!", LCD_WIDTH + 1 );
    menu_display_messagebox( );
  }
}

// #####################################################################

void sd_routine( )
{
  // check timer
  if( current_millis < sd_routine_millis )
    return;
  sd_routine_millis = current_millis + 5000;

  if( b_sd_initialized == false )
  {
    // we'll use the initialization code from the utility libraries
    // since we're just testing if the card is working!
    if( !card.init( SPI_HALF_SPEED, chipSelect ) )
    {
      sd_status = SD_STATUS_NO_CARD;
      menu_sd[0].data = (void *) "No Card";
      Serial.println(F("[SD] Failed to establish connection with SD card."));
      return;
    }

    switch( card.type( ) )
    {
      case SD_CARD_TYPE_SD1:  menu_sd[1].data = (void *) "SD1"; break;
      case SD_CARD_TYPE_SD2:  menu_sd[1].data = (void *) "SD2"; break;
      case SD_CARD_TYPE_SDHC: menu_sd[1].data = (void *) "SDHC"; break;
      default:                menu_sd[1].data = (void *) "Unknown"; break;
    }
    
    // Now we will try to open the 'volume'/'partition' - it should be FAT16 or FAT32
    if( !volume.init( card ) )
    {
      sd_status = SD_STATUS_NO_FILESYSTEM;
      menu_sd[0].data = (void *) "No FS";
      Serial.println(F("[SD] Could not find FAT16/FAT32 partition. Make sure you've formatted the card."));
      return;
    }

    switch( volume.fatType( ) )
    {
      case 16: menu_sd[2].data = (void *) "FAT16"; break;
      case 32: menu_sd[2].data = (void *) "FAT32"; break;
      default: menu_sd[2].data = (void *) "Unknown"; break;
    }

    volumesize = volume.blocksPerCluster();    // clusters are collections of blocks
    volumesize *= volume.clusterCount();       // we'll have a lot of clusters
    volumesize /= 2;                           // SD card blocks are always 512 bytes (2 blocks are 1KB)
    snprintf( volumesize_txt, sizeof(volumesize_txt), "%.1f MiB", ((float) volumesize) / 1024.0 );
    menu_sd[3].data = &volumesize_txt;

    // Card info retreived, now initialize it
    if ( !SD.begin( chipSelect ) )
    {
      sd_status = SD_STATUS_NO_CARD;
      menu_sd[0].data = (void *) "No Card";
      Serial.println(F("[SD] Card failed, or not present."));
      return;
    }

    menu_sd[0].data = (void *) "OK";
    sd_status = SD_STATUS_OK;
    Serial.println(F("[SD] SD card subsystem initialized."));
    b_sd_initialized = true;

    // The card is ready. Read the configuration file
    sd_read_configuration_command( NULL );
  }
  else
  {
    
  }
}

// #####################################################################
